#include "PCA301Plug.h"
#include "Arduino.h"

#define PROGNAME         "LaCrosseITPlusReader.Gateway"
#define PROGVERS         "1.15"

#define RFM1_SS          15
#define RFM2_SS          2
#define RFM3_SS          0
#define LED_PIN          16

// Arduino / ESP8266 libs
#include "SPI.h"
#include "Wire.h"
#include "EEPROM.h"
#include "ESP8266WiFi.h"
#include "WiFiClient.h"
#include "ESP8266WebServer.h"
#include "pins_arduino.h"
#include "ESP8266HTTPClient.h"
#include "ESP8266httpUpdate.h"
#include "WiFiUdp.h"
#include "ESP8266mDNS.h"
#include "ArduinoOTA.h"


// Other libs
#include "BMP180.h"
#include "RFMxx.h"
#include "SensorBase.h"
#include "LaCrosse.h"
#include "EMT7110.h"
#include "TX38IT.h"
#include "WSBase.h"
#include "WS1080.h"
#include "TX22IT.h"
#include "WT440XH.h"
#include "CustomSensor.h"
#include "LevelSenderLib.h"
#include "ESP8266Tools.h"
#include "TypedQueue.h"
#include "HashMap.h"
#include "HTML.h"
#include "DHTxx.h"
#include "I2CBase.h"
#include "BME280.h"
#include "LM75.h"
#include "EC3000.h"
#include "ESP8266OTA.h"

// In this sketch
#include "AccessPoint.h"
#include "WebFrontend.h"
#include "DataPort.h"
#include "Settings.h"
#include "StateManager.h"
#include "OTAUpdate.h"
#include "OwnSensors.h"
#include "PCA301Plug.h"
#include "PCA301.h"
#include "PCA301PlugList.h"

extern "C" {
#include "ets_sys.h"
#include "os_type.h"
#include "osapi.h"
#include "mem.h"
#include "user_interface.h"
#include "cont.h"
}

// The following settings can also be set from FHEM
#define ENABLE_ACTIVITY_LED    1         // <n>a       set to 0 if the blue LED bothers
unsigned long DATA_RATE_S1   = 17241ul;  // <n>c       use one of the possible data rates (for transmit on RFM #1)
bool DEBUG                   = 0;        // <n>d       set to 1 to see debug messages
                                         // <1>e       Clear EEPROM
unsigned long INITIAL_FREQ   = 868300;   // <n>f       initial frequency in kHz (5 kHz steps, 860480 ... 879515) 
int ALTITUDE_ABOVE_SEA_LEVEL = 0;        // <n>h       Altitude
                                         // <n,f,i>i   Init PCA for Radio #<n> to <m>MHz and <i>s Interval
byte TOGGLE_MODE_R1          = 3;        // <n>m       bits 1: 17.241 kbps, 2 : 9.579 kbps, 4 : 8.842 kbps, 8 : 20.000 kbps (for RFM #1)
byte TOGGLE_MODE_R2          = 3;        // <n>M       bits 1: 17.241 kbps, 2 : 9.579 kbps, 4 : 8.842 kbps, 8 : 20.000 kbps (for RFM #2)
byte TOGGLE_MODE_R3          = 3;        // <n>M       bits 1: 17.241 kbps, 2 : 9.579 kbps, 4 : 8.842 kbps, 8 : 20.000 kbps (for RFM #3)
                                         // <n>o       set HF-parameter e.g. 50305o for RFM12 or 1,4o for RFM69
byte PASS_PAYLOAD            = 0;        // <n>p       transmitted the payload on the serial port 1: all, 2: only undecoded data
unsigned long DATA_RATE_R1   = 17241ul;  // <n>r       use one of the possible data rates (for RFM #1)
unsigned long DATA_RATE_R2   = 9579ul;   // <n>R       use one of the possible data rates (for RFM #2)
unsigned long DATA_RATE_R3   = 8842ul;   // <n>R       use one of the possible data rates (for RFM #3)

                                         // <x,x,...>s Send to PCA301 (must be 10 byte)
                                         // <x,x,...>S Send to CustomSensor

uint16_t TOGGLE_INTERVAL_R1  = 0;        // <n>t       0=no toggle, else interval in seconds (for RFM #1)
uint16_t TOGGLE_INTERVAL_R2  = 0;        // <n>T       0=no toggle, else interval in seconds (for RFM #2)
uint16_t TOGGLE_INTERVAL_R3  = 0;        // <n>T       0=no toggle, else interval in seconds (for RFM #3)

                                         // v          show version
bool USE_WIFI                = 1;        // <n>w       0=no wifi
                                         // x          test command 
bool ANALYZE_FRAMES          = 0;        // <n>z       set to 1 to display analyzed frame data instead of the normal data
uint KVINTERVAL              = 10;       //            Interval for KV-transmision

bool USE_SERIAL              = 1;        //            0=do not send sensor data on the serial
bool DEBUG_PCA301            = 0;        //            debug PCA301
bool DEBUG_ESP               = 0;        //            set to 1 to see ESP8266 related debug messages

// wifi settings
#define FRONTEND_PORT          80        // Port for the Web frontend
#define DATA_PORT1             81        // Port for data
#define DATA_PORT2             82        // Port for data
#define DATA_PORT3             83        // Port for data
#define OTA_PORT               8266
IPAddress AP_IP              = IPAddress(192, 168, 222, 1);
IPAddress AP_SUBNET          = IPAddress (255, 255, 225, 0);

// --- Variables -------------------------------------------------------------------------------------------------------
unsigned long lastToggleR1 = 0;
unsigned long lastToggleR2 = 0;
unsigned long lastToggleR3 = 0;
unsigned long commandData[32];
byte commandDataPointer = 0;
ESP8266Tools esp(LED_PIN);
RFMxx rfm1(13, 12, 14, RFM1_SS);
RFMxx rfm2(13, 12, 14, RFM2_SS);
RFMxx rfm3(13, 12, 14, RFM3_SS);
OwnSensors ownSensors;
WebFrontend frontend(FRONTEND_PORT);
DataPort dataPort1;
DataPort dataPort2;
DataPort dataPort3;
AccessPoint accessPoint (AP_IP, AP_IP, AP_SUBNET, "LaCrosseGateway");
StateManager stateManager;
PCA301 pca301;
ESP8266OTA ota;
unsigned int lastOtaProgress;

void Dispatch(String data) {
  if(USE_SERIAL) {
    Serial.println(data);
  }
  
  if (USE_WIFI) {
    dataPort1.AddPayload(data);
    dataPort2.AddPayload(data);
    dataPort3.AddPayload(data);
    esp.Blink(1);
  }
}

void SetDataRate(RFMxx *rfm, unsigned long dataRate) {
  if(rfm->GetDataRate() != 20000 && dataRate == 20000) {
    rfm->InitializeEC3000();
    rfm->EnableReceiver(true);
  }
  else if(rfm->GetDataRate() == 20000 && dataRate != 20000) {
    rfm->InitializeLaCrosse();
    rfm->EnableReceiver(true);
  }
  

  rfm->SetDataRate(dataRate);
}

static void HandleSerialPort(char c) {
  static unsigned long value;
  static bool rfmNumberFlag;
  static byte rfmNumber = 0;
  unsigned long dataRate = 0;
  if (c == ',') {
    commandData[commandDataPointer++] = value;
    value = 0;
  }
  else if(rfmNumberFlag) {
    rfmNumber = c - '0';
    rfmNumberFlag = false;
  }
  else if(c == '#') {
    rfmNumberFlag = true;
  }
  else if ('0' <= c && c <= '9') {
    value = 10 * value + c - '0';
  }
  else if (('a' <= c && c <= 'z') || ('A' <= c && c <= 'Z')) {
    if (rfmNumber == 0) {
      rfmNumber = ('A' <= c && c <= 'Z') ? 2 : 1;
    }
  
    switch (c) {
    case 'd':
      // DEBUG
      SetDebugMode(value, false);
      break;
    case 'h':
      // height
      ownSensors.SetAltitudeAboveSeaLevel(value);
      break;
    case 'x':
    case 'X':
      // Tests
      HandleCommandX(value);
      break;
    case 'a':
      // Activity LED    
      esp.EnableLED(value);
      break;
    case 'w':
      // wifi    
      if (!value) {
        StopWifi();
      }
      break;
    case 'r':
    case 'R':
      // Data rate
      switch (value) {
      case 0:
        dataRate = 17241ul;
        break;
      case 1:
        dataRate = 9579ul;
        break;
      case 2:
        dataRate = 8842ul;
        break;
      case 3:
        dataRate = 20000ul;
        break;
      default:
        dataRate = value;
        break;
      }
      if (rfmNumber == 1 && rfm1.IsConnected()) {
        DATA_RATE_R1 = dataRate;
        SetDataRate(&rfm1, DATA_RATE_R1);
      }
      else if (rfmNumber == 2 && rfm2.IsConnected()) {
        DATA_RATE_R2 = dataRate;
        SetDataRate(&rfm2, DATA_RATE_R2);
      }
      else if (rfmNumber == 3 && rfm3.IsConnected()) {
        DATA_RATE_R3 = dataRate;
        SetDataRate(&rfm3, DATA_RATE_R3);
      }
      break;
    case 'm':
    case 'M':
      if (rfmNumber == 1 && rfm1.IsConnected()) {
        TOGGLE_MODE_R1 = value;
      }
      else if (rfmNumber == 2 && rfm2.IsConnected()) {
        TOGGLE_MODE_R2 = value;
      }
      else if (rfmNumber == 3 && rfm3.IsConnected()) {
        TOGGLE_MODE_R3 = value;
      }
      break;
    case 'o':
    case 'O':
      // Set HF parameter
      commandData[commandDataPointer] = value;
      HandleCommandO(rfmNumber, value, commandData, ++commandDataPointer);
      commandDataPointer = 0;
      break;
    case 'p':
      PASS_PAYLOAD = value;
      break;
    case 's':
      // Send
      commandData[commandDataPointer] = value;
      HandlePCA301Send(commandData, ++commandDataPointer);
      commandDataPointer = 0;
      break;
    case 'S':
      // Send
      commandData[commandDataPointer] = value;
      HandleCommandS(commandData, ++commandDataPointer);
      commandDataPointer = 0;
      break;
    case 'i':
      // Init PCA301
      commandData[commandDataPointer] = value;
      HandleCommandI(commandData, ++commandDataPointer);
      commandDataPointer = 0;
      break;
    case 't':
    case 'T':
      // Toggle data rate
      if (rfmNumber == 1 && rfm1.IsConnected()) {
        TOGGLE_INTERVAL_R1 = value;
      }
      else if (rfmNumber == 2 && rfm2.IsConnected()) {
        TOGGLE_INTERVAL_R2 = value;
      }
      if (rfmNumber == 3 && rfm3.IsConnected()) {
        TOGGLE_INTERVAL_R3 = value;
      }
      break;
    case 'v':
      // Version info
      HandleCommandV();
      break;

    case 'f':
    case 'F':
      if (rfmNumber == 1 && rfm1.IsConnected()) {
        rfm1.SetFrequency(value); 
      }
      else  if (rfmNumber == 2 && rfm2.IsConnected()) {
        rfm2.SetFrequency(value);
      }
      else  if (rfmNumber == 3 && rfm3.IsConnected()) {
        rfm3.SetFrequency(value);
      }
      break;

    case 'z':
      ANALYZE_FRAMES = value;
      break;
    
    case 'e':
      if (value == 8266) {
        EEPROM.begin(EEPROM_SIZE);
        for (int i = 0; i < EEPROM_SIZE; i++) {
          EEPROM.write(i, 0);
        }
        EEPROM.commit();
        EEPROM.end();
        delay(500);
        ESP.reset();
        break;
      }

    default:
      HandleCommandV();
      break;
    }
    value = 0;
    rfmNumberFlag = false;
    rfmNumber = 0;
  }
  else if (' ' < c && c < 'A') {
    HandleCommandV();
  }
}

void SetDebugMode(boolean mode, boolean modeESP) {
  DEBUG = mode;
  SensorBase::SetDebugMode(mode);
  rfm1.SetDebugMode(mode);
  rfm2.SetDebugMode(mode);
  rfm3.SetDebugMode(mode);
  pca301.SetDebugMode(DEBUG_PCA301);

  dataPort1.SetDebugMode(modeESP);
  dataPort2.SetDebugMode(modeESP);
  dataPort3.SetDebugMode(modeESP);
  frontend.SetDebugMode(modeESP);
  accessPoint.SetDebugMode(modeESP);
  Settings::SetDebugMode(modeESP);
  
  
}


void HandleCommandI(unsigned long *commandData, byte length){
  word interval = 60;
  if (length == 3) {
    interval = commandData[2];
  }

  if (length >= 2) {
    RFMxx *rfm;
    switch (commandData[0]) {
    case 1:
      rfm = &rfm1;
      break;
    case 2:
      rfm = &rfm2;
      break;
    case 3:
      rfm = &rfm3;
      break;
    default:
      rfm = NULL;
      break;
    }
    if (rfm != NULL){
      pca301.Begin(rfm, commandData[1], interval);
    }
  }

}


// This function is for testing 
void HandleCommandX(byte value) {

  while (1);

  ////byte command = 5;

  ////if (value == 2) {
  ////  value = 0;
  ////  command = 4;
  ////}
  ////
  ////byte payload[10];
  ////payload[0] = 1;
  ////payload[1] = 5;
  ////payload[2] = 0x03;
  ////payload[3] = 0xA0;
  ////payload[4] = 0x94;
  ////payload[5] = value;
  ////payload[6] = 0xFF;
  ////payload[7] = 0xFF;
  ////payload[8] = 0xFF;
  ////payload[9] = 0xFF;
  ////pca301.SendPayload(payload, false);
 
}

void HandleCommandS(unsigned long *data, byte size) {
  rfm1.EnableReceiver(false);

  struct CustomSensor::Frame frame;
  frame.ID = data[0];
  frame.NbrOfDataBytes = size - 1;

  for (int i = 0; i < frame.NbrOfDataBytes; i++) {
    frame.Data[i] = data[i + 1];
  }

  CustomSensor::SendFrame(&frame, &rfm1, DATA_RATE_S1);

  rfm1.EnableReceiver(true);
}

void HandlePCA301Send(unsigned long *data, byte size) {
  if (size == 10 && pca301.IsInitialized()) {
    rfm1.EnableReceiver(false);

    byte payload[10];
    for (int i = 0; i < 10; i++) {
      payload[i] = data[i];
    }
    pca301.SendPayload(payload, false);

    rfm1.EnableReceiver(true);
  }
}


void HandleCommandO(byte rfmNbr, unsigned long value, unsigned long *data, byte size) {
  // 50305o (is 0xC481) for RFM12 or 1,4o for RFM69
  if (size == 1 && rfm1.GetRadioType() == RFMxx::RFM12B) {
    if (rfmNbr == 1) {
      rfm1.SetHFParameter(value);
    }
    else if (rfmNbr == 2) {
      rfm2.SetHFParameter(value);
    }
    else if (rfmNbr == 3) {
      rfm3.SetHFParameter(value);
    }
  }
  else if (size == 2 && rfm1.GetRadioType() == RFMxx::RFM69CW) {
    if (rfmNbr == 1) {
      rfm1.SetHFParameter(data[0], data[1]);
    }
    else if (rfmNbr == 2) {
      rfm2.SetHFParameter(data[0], data[1]);
    }
    else if (rfmNbr == 3) {
      rfm3.SetHFParameter(data[0], data[1]);
    }
  }
}

void HandleCommandV() {
  String result = "";
  
  result += "\r\n[";
  result += PROGNAME;
  result += ".";
  result += PROGVERS;

  if (rfm1.IsConnected()) {
    result += " (1=";
    result += rfm1.GetRadioName();
    result += " f:";
    result += rfm1.GetFrequency();

    if (TOGGLE_INTERVAL_R1) {
      result += " t:";
      result += TOGGLE_INTERVAL_R1;
      result += "~";
      result += TOGGLE_MODE_R1;
    }
    else {
      result += " r:";
      result += rfm1.GetDataRate();
    }
    result += ")";
  }

  if (rfm2.IsConnected()) {
    result += " + (2=";
    result += rfm2.GetRadioName();
    result += " f:";
    result += rfm2.GetFrequency();
    if (TOGGLE_INTERVAL_R2) {
      result += " t:";
      result += TOGGLE_INTERVAL_R2;
      result += "~";
      result += TOGGLE_MODE_R2;
    }
    else {
      result += " r:";
      result += rfm2.GetDataRate();

    }
    result += ")";
  }

  if (rfm3.IsConnected()) {
    result += " + (3=";
    result += rfm3.GetRadioName();
    result += " f:";
    result += rfm3.GetFrequency();
    if (TOGGLE_INTERVAL_R3) {
      result += " t:";
      result += TOGGLE_INTERVAL_R3;
      result += "~";
      result += TOGGLE_MODE_R3;
    }
    else {
      result += " r:";
      result += rfm3.GetDataRate();

    }
    result += ")";
  }

  if (ownSensors.HasBMP180()) {
    result += " + BMP180";
  }
  if (ownSensors.HasBME280()) {
    result += " + BME280";
  }
  if (ownSensors.HasDHT22()) {
    result += " + DHT22";
  }
  if (ownSensors.HasLM75()) {
    result += " + LM75";
  }

  result += " {IP=";
  if (WiFi.status() == WL_CONNECTED) {
    result += WiFi.localIP().toString();
  }
  else {
    result += WiFi.softAPIP().toString();
  }
  result += "}";

  result += "]";
  
  Dispatch(result);

}

bool HandleReceivedData(RFMxx *rfm) {
  bool result = false;
  
  rfm->EnableReceiver(false);

  byte payload[PAYLOADSIZE];
  rfm->GetPayload(payload);

  rfm->EnableReceiver(true);

  if (ANALYZE_FRAMES) {
    ////WS1080::AnalyzeFrame(payload);
    ////TX22IT::AnalyzeFrame(payload);
    ////LaCrosse::AnalyzeFrame(payload);
    ////LevelSenderLib::AnalyzeFrame(payload);
    ////EMT7110::AnalyzeFrame(payload);
    ////TX38IT::AnalyzeFrame(payload);
    ////CustomSensor::AnalyzeFrame(payload);
    EC3000::AnalyzeFrame(payload);
    ////Serial.println();
  }
  else if (PASS_PAYLOAD == 1) {
    for (int i = 0; i < PAYLOADSIZE; i++) {
      Serial.print(payload[i], HEX);
      Serial.print(" ");
    }
    Serial.println();
  }
  else {
    if (DEBUG) {
      Serial.print("\r\nEnd receiving, HEX raw data: ");
      for (int i = 0; i < 16; i++) {
        Serial.print(payload[i], HEX);
        Serial.print(" ");
      }
      Serial.println();
    }

    String data = "";

    // Try LaCrosse like TX29DTH
    if (data.length() == 0 && LaCrosse::IsValidDataRate(rfm->GetDataRate())) {
      data = LaCrosse::GetFhemDataString(payload);
    }
    
    // Try EC3000
    if (data.length() == 0 && EC3000::IsValidDataRate(rfm->GetDataRate())) {
      data = EC3000::GetFhemDataString(payload);
    }

    // Try TX22IT (WS 1600)
    if (data.length() == 0 && TX22IT::IsValidDataRate(rfm->GetDataRate())) {
      data = TX22IT::GetFhemDataString(payload);
    }

    // Try WS 1080
    if (data.length() == 0 && WS1080::IsValidDataRate(rfm->GetDataRate())) {
      data = WS1080::GetFhemDataString(payload);
    }

    // Try LevelSender
    if (data.length() == 0 && LevelSenderLib::IsValidDataRate(rfm->GetDataRate())) {
      data = LevelSenderLib::GetFhemDataString(payload);
    }

    // Try EMT7110
    if (data.length() == 0 && EMT7110::IsValidDataRate(rfm->GetDataRate())) {
      data = EMT7110::GetFhemDataString(payload);
    }

    // Try TX38IT
    if (data.length() == 0 && TX38IT::IsValidDataRate(rfm->GetDataRate())) {
      data = TX38IT::GetFhemDataString(payload);
    }

    // Try WT440XH
    if (data.length() == 0 && WT440XH::IsValidDataRate(rfm->GetDataRate())) {
      data = WT440XH::GetFhemDataString(payload);
    }

    // Try CustomSensor
    if (data.length() == 0 && CustomSensor::IsValidDataRate(rfm->GetDataRate())) {
      data = CustomSensor::GetFhemDataString(payload);
    }
    
    // Try PCA301
    if (pca301.IsInitialized() && data.length() == 0 && PCA301::IsValidDataRate(rfm->GetDataRate())) {
      data = pca301.GetFhemDataString(payload);
    }

    if (data.length() > 0) {
      result = true;
      Dispatch(data);
    }
  }

  
  
  return result;
}

void HandleDataRateToggle(RFMxx *rfm, unsigned long *lastToggle, unsigned long *dataRate, uint16_t interval, byte toggleMode) {
  if (interval > 0) {
    // After about 50 days millis() will overflow to zero 
    if (millis() < *lastToggle) {
      *lastToggle = 0;
    }

    if (millis() > *lastToggle + interval * 1000 && toggleMode > 0) {
      // Bits 1: 17.241 kbps, 2 : 9.579 kbps, 4 : 8.842 kbps, 8 : 20.000 kbps

      HashMap<unsigned long, unsigned long, 4> dataRates;
      if (toggleMode & 1) {
        dataRates.Put(17241, 17241);
      }
      if (toggleMode & 2) {
        dataRates.Put(9579, 9579);
        if (dataRates.Size() > 0) {
          *dataRates.GetValuePointerAt(dataRates.Size() - 2) = 9579;
        }
      }
      if (toggleMode & 4) {
        dataRates.Put(8842, 8842);
        if (dataRates.Size() > 0) {
          *dataRates.GetValuePointerAt(dataRates.Size() - 2) = 8842;
        }
      }
      if (toggleMode & 8) {
        dataRates.Put(20000, 20000);
        if (dataRates.Size() > 0) {
          *dataRates.GetValuePointerAt(dataRates.Size() - 2) = 20000;
        }
      }
      *dataRates.GetValuePointerAt(dataRates.Size() - 1) = dataRates.GetKeyAt(0);

      *dataRate = dataRates.Get(rfm->GetDataRate(), 0);
      if (*dataRate == 0) {
        *dataRate = 17241;
      }

      SetDataRate(rfm, *dataRate);
      *lastToggle = millis();

    }
  }
}

static bool StartWifi(Settings settings) {
  bool result = false;

  if (DEBUG_ESP) {
    Serial.println("Disconnect wifi");
  }
  WiFi.disconnect();

  if (DEBUG_ESP) {
    Serial.println("Start WIFI_STA");
  }
  WiFi.mode(WiFiMode::WIFI_STA);

  String hostName = settings.Get("HostName", "LaCrosseGateway");
  if (DEBUG_ESP) {
    Serial.printf("HostName is: %s\r\n", hostName.c_str());
  }
  WiFi.hostname(hostName);
  
  String staticIP = settings.Get("staticIP", "");
  String staticMask = settings.Get("staticMask", "");
  String staticGW = settings.Get("staticGW", "");

  if (staticIP.length() < 7 || staticGW.length() < 7 || staticMask.length() < 7){
    if (DEBUG_ESP) {
      Serial.println("Using DHCP");
    }
  }
  else {
    if (DEBUG_ESP) {
      Serial.println("Using static IP");
      Serial.printf("IP: %s\r\n", staticIP.c_str());
      Serial.printf("Mask: %s\r\n", staticMask.c_str());
      Serial.printf("Gateway: %s\r\n", staticGW.c_str());
    }
    WiFi.config(HTML::IPAddressFromString(staticIP),
                HTML::IPAddressFromString(staticGW),
                HTML::IPAddressFromString(staticMask));
  }

  WiFi.begin(settings.Get("ctSSID", "---").c_str(), settings.Get("ctPASS", "---").c_str());

  if (DEBUG_ESP) {
    Serial.println("Trying to connect to an AP");
  }
  
  esp.SwitchLed(true, true);
  byte retryCounter = 0;
  while (retryCounter < 30 && WiFi.status() != WL_CONNECTED) {
    retryCounter++;
    delay(500);
    if (DEBUG_ESP) {
      Serial.print(".");
    }
    esp.SwitchLed(retryCounter % 2 == 0, true);
  }
  esp.SwitchLed(false, true);

  if (WiFi.status() == WL_CONNECTED) {
    result = true;
    if (DEBUG_ESP) {
      Serial.println();
      Serial.println("connected :-)");
      Serial.print("SSID: ");
      Serial.println(WiFi.SSID());
      Serial.print("IP: ");
      Serial.println(WiFi.localIP());
    }

    if (settings.Get("Flags", "").indexOf("MDNS") != - 1) {
      if (DEBUG_ESP) {
        Serial.println("Starting MDNS");
      }
      MDNS.begin("esp8266-ota", WiFi.localIP());
      MDNS.addService("arduino", "tcp", OTA_PORT);
      MDNS.addService("http", "tcp", FRONTEND_PORT);
    }

  }
  else {
    if (DEBUG_ESP) {
      Serial.println();
      Serial.println("We got no connection :-(");
    }
    // Open access point for 15 minutes
    accessPoint.Begin(900);
    esp.SwitchLed(true, true);
  }

  frontend.Begin(&stateManager);
  ota.Begin(frontend.WebServer());

  
  uint p1 = settings.GetInt("DataPort1", 81);
  uint p2 = settings.GetInt("DataPort2", 0);
  uint p3 = settings.GetInt("DataPort3", 0);
  if (p1 > 0) {
    dataPort1.Begin(p1);
  }
  if (p2 > 0) {
    dataPort2.Begin(p2);
  }
  if (p3 > 0) {
    dataPort3.Begin(p3);
  }

  USE_WIFI = 1;
  
  return result;
}

static void StopWifi() {
  WiFi.mode(WiFiMode::WIFI_OFF);
  USE_WIFI = 0;
}

////void OnSubProcessorReceive(String data) {
////  Serial.println(data);
////}

void setup(void) {
  Serial.begin(57600);
  delay(1000);
  Serial.println();
  if (DEBUG_ESP) {
    Serial.println("This is line 5 of void setup(void)");
    Serial.print(PROGNAME);
    Serial.print(" V");
    Serial.println(PROGVERS);
    Serial.print("Free heap: ");
    Serial.println(ESP.getFreeHeap());

  }
 
  SetDebugMode(DEBUG, DEBUG_ESP);

  rfm1.Begin();
  rfm2.Begin();
  rfm3.Begin();

  Settings settings;
  settings.Read();
  KVINTERVAL = settings.GetInt("KVInterval", KVINTERVAL);

  esp.EnableLED(ENABLE_ACTIVITY_LED);
  lastToggleR1 = millis();
  lastToggleR2 = millis();

  esp.Blink(5, true);
  
  // Start wifi
  if (USE_WIFI) {
    if (DEBUG_ESP) {
      Serial.println("Read settings");
    }

    if (DEBUG_ESP) {
      Serial.println("Settings are:");
      settings.Dump();
    }
  
    if (DEBUG_ESP) {
      Serial.println("Starting wifi");
    }
    StartWifi(settings);

    if (WiFi.status() == WL_CONNECTED) {
      ArduinoOTA.onStart([]() {
        lastOtaProgress = 0;
        if (DEBUG_ESP) {
          Serial.println("Starting OTA update");
        }
      });
      ArduinoOTA.onEnd([]() {
        if (DEBUG_ESP) {
          Serial.println();
          Serial.println("OTA update finished");
        }
      });
      
      ArduinoOTA.onProgress([](unsigned int progress, unsigned int total) {
        if (progress / (total / 100) == lastOtaProgress + 2) {
          Serial.print(".");
          lastOtaProgress +=2;
        }
      });
      ArduinoOTA.onError([](ota_error_t error) {
        if (DEBUG_ESP) {
          Serial.printf("Error[%u]: ", error);
          if (error == OTA_AUTH_ERROR) Serial.println("Auth Failed");
          else if (error == OTA_BEGIN_ERROR) Serial.println("Begin Failed");
          else if (error == OTA_CONNECT_ERROR) Serial.println("Connect Failed");
          else if (error == OTA_RECEIVE_ERROR) Serial.println("Receive Failed");
          else if (error == OTA_END_ERROR) Serial.println("End Failed");
        }
      });
      ArduinoOTA.setPort(OTA_PORT);
      ArduinoOTA.begin();
      if (DEBUG_ESP) {
        Serial.print("OTA running on port ");
        Serial.println(OTA_PORT);
      }


    }

  }

  if (DEBUG_ESP) {
    Serial.println("Searching RFMs and BMP");
  }
  if (rfm1.IsConnected()) {
    rfm1.InitializeLaCrosse();
    rfm1.SetFrequency(INITIAL_FREQ);
    SetDataRate(&rfm1, DATA_RATE_R1);
    rfm1.EnableReceiver(true);
    if (DEBUG_ESP) {
      Serial.print("First RFM found: ");
      Serial.println(rfm1.GetRadioName());
    }
  }

  if (rfm2.IsConnected()) {
    rfm2.InitializeLaCrosse();
    rfm2.SetFrequency(INITIAL_FREQ);
    SetDataRate(&rfm2, DATA_RATE_R2);
    rfm2.EnableReceiver(true);
    if (DEBUG_ESP) {
      Serial.print("Second RFM found: ");
      Serial.println(rfm2.GetRadioName());
    }
  }

  if (rfm3.IsConnected()) {
    rfm3.InitializeLaCrosse();
    rfm3.SetFrequency(INITIAL_FREQ);
    SetDataRate(&rfm3, DATA_RATE_R3);
    rfm3.EnableReceiver(true);
    if (DEBUG_ESP) {
      Serial.print("Third RFM found: ");
      Serial.println(rfm3.GetRadioName());
    }
  }
 
  Wire.begin(0x11);
  ownSensors.SetAltitudeAboveSeaLevel(ALTITUDE_ABOVE_SEA_LEVEL);
  ownSensors.TryInitialize();
  if (DEBUG_ESP) {
    if (ownSensors.HasBMP180()) {
      Serial.println("BMP180 found");
    }
    if (ownSensors.HasBME280()) {
      Serial.println("BME280 found");
    }
    if (ownSensors.HasDHT22()) {
      Serial.println("DHT22 found");
    }
    if (ownSensors.HasLM75()) {
      Serial.println("LM75 found");
    }
  }
  
  stateManager.Begin(PROGVERS);

  // FHEM needs this information
  delay(1000);
  if (DEBUG_ESP) {
    Serial.println("Sending init String to FHEM");
  }
  HandleCommandV();

  if (DEBUG_ESP) {
    Serial.println("Setup completely done");
  }

}

String CommandHandler(String command) {
  for (uint i = 0; i < command.length(); i++) {
    HandleSerialPort(command.charAt(i));
  }
  stateManager.ResetLastFullKVPUpdate();
  return "";
}


// **********************************************************************
void loop(void) {
  // Handle the commands from the serial port
  // ----------------------------------------
  if (Serial.available()) {
    HandleSerialPort(Serial.read());
  }

  // Periodically send own sensor data
  // ---------------------------------
  String data = ownSensors.GetFhemDataString();
  if (data.length() > 0) {
    Dispatch(data);
  }
  
  // Handle the data reception
  // -------------------------
  byte receivedPackets = 0;
  if (rfm1.IsConnected()) {
    rfm1.Receive(); 
    if (rfm1.PayloadIsReady()) {
      if (HandleReceivedData(&rfm1)) {
        receivedPackets++;
      }
    }
  }
  if (rfm2.IsConnected()) {
    rfm2.Receive();
    if (rfm2.PayloadIsReady()) {
      if(HandleReceivedData(&rfm2)) {
        receivedPackets++;
      }
    }
  }
  if (rfm3.IsConnected()) {
    rfm3.Receive();
    if (rfm3.PayloadIsReady()) {
      if (HandleReceivedData(&rfm3)) {
        receivedPackets++;
      }
    }
  }
  
  // Handle PCA301
  if(pca301.IsInitialized()) {
    pca301.Handle();
  }

  // Periodically send some info about us
  // ------------------------------------
  stateManager.Handle(receivedPackets);
  if (KVINTERVAL > 1) {
    String kvData = stateManager.GetKVP(KVINTERVAL);
    if (kvData.length() > 0) {
      Dispatch(kvData);
    }
  }

  // Handle the data rate
  // --------------------
  if (rfm1.IsConnected()) {
    HandleDataRateToggle(&rfm1, &lastToggleR1, &DATA_RATE_R1, TOGGLE_INTERVAL_R1, TOGGLE_MODE_R1);
  }
  if (rfm2.IsConnected()) {
    HandleDataRateToggle(&rfm2, &lastToggleR2, &DATA_RATE_R2, TOGGLE_INTERVAL_R2, TOGGLE_MODE_R2);
  }
  if (rfm3.IsConnected()) {
    HandleDataRateToggle(&rfm3, &lastToggleR3, &DATA_RATE_R3, TOGGLE_INTERVAL_R3, TOGGLE_MODE_R3);
  }

  // Handle the web frontend and the access point
  // --------------------------------------------
  if(USE_WIFI) {
    frontend.Handle();
    accessPoint.Handle();
  }

  // Handle the port for FHEM
  // ------------------------
  if (USE_WIFI) {
    if (dataPort1.Handle(CommandHandler) || dataPort2.Handle(CommandHandler) || dataPort3.Handle(CommandHandler) ) {
      ////esp.Blink(1);
    }
  }
  
  // Handle OTA
  // ----------
  if (WiFi.status() == WL_CONNECTED) {
    ArduinoOTA.handle();
  }
}
